@@ -41,4 +41,4 @@ del get_versions
 
 
 setattr(Tensor, "clip", clip)
-prepare_op = Tensor._op
+execute_op = Tensor._op
